#define _UNUSED(n)

//#include "glog.h"
//#define IM_ASSERT( exp ) \
//    ( (exp) ? (void)0 : glog_v("Failed assertion at %s:%d %s\n",  __FILE__,__LINE__, #exp))

#include "gplugin.h"
#include "gfile.h"
#include "gstdio.h"
#include "ginput.h"

#include "application.h"
#include "luaapplication.h"
#include "lua.hpp"
#include "luautil.h"
#include "sprite.h"
#include "binder.h"

#include "texturebase.h"
#include "bitmapdata.h"

#include "imgui_src/imgui.h"

////////////////////////////////////////////////////////////////////////////////
///
/// HELPERS
///
////////////////////////////////////////////////////////////////////////////////

struct MyTextureData
{
    void *texture;
    float width;
    float height;
};

double getfield(lua_State *L, const char *key)
{

    lua_pushstring(L, key);
    lua_gettable(L, -2);  // get table[key]
    double result = lua_tonumber(L, -1);
    lua_pop(L, 1);  // remove number
    return result;
}

int getKeyboardModifiers(lua_State *L)
{
    lua_getglobal(L, "application");
    lua_getfield(L, -1, "getKeyboardModifiers");
    lua_getglobal(L, "application");
    lua_call(L,1,1);
    int mod = luaL_checkinteger(L, -1);
    lua_pop(L, 1);

    return mod;
}

MyTextureData getTexture(lua_State *L, int idx = 1)
{
    Binder binder(L);

    if (binder.isInstanceOf("TextureBase", idx))
    {
        MyTextureData data;
        TextureBase* textureBase = static_cast<TextureBase*>(binder.getInstance("TextureBase", idx));
        //textureBase->data->
        data.texture = (void *)textureBase->data->gid;
        data.width = (float)textureBase->data->width / (float)textureBase->data->exwidth;
        data.height = (float)textureBase->data->width / (float)textureBase->data->exheight;
        return data;
    }
    else if (binder.isInstanceOf("TextureRegion", idx))
    {
        MyTextureData data;
        BitmapData* bitmapData = static_cast<BitmapData*>(binder.getInstance("TextureRegion", idx));
        //return (void *)bitmapData->texture()->data->gid;
        data.texture = (void *)bitmapData->texture()->data->gid;
        data.width = (float)bitmapData->texture()->data->width / (float)bitmapData->texture()->data->exwidth;
        data.height = (float)bitmapData->texture()->data->width / (float)bitmapData->texture()->data->exheight;
        return data;
    }
    else
    {
        lua_pushstring(L, "Type mismatch. 'TextureBase' or 'TextureRegion' expected.");
        lua_error(L);
    }
}

ImVec4 getColor(lua_State *L, int idx, ImU32 def = 0)
{ 
    ImVec4 color = ImGui::ColorConvertU32ToFloat4(luaL_optinteger(L, idx, def));

    float tmp = color.x;
    color.x = color.w;
    color.w = tmp;

    tmp = color.y;
    color.y = color.z;
    color.z = tmp;

    return color;
}

void luaError(lua_State *L, const char *message)
{
    lua_pushstring(L, message);
    lua_error(L);
}


/////////////////////////////////////////////////////////////////////////////////////////////
///
/// ENUMS
///
/////////////////////////////////////////////////////////////////////////////////////////////

void bindWindowFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "WindowFlags_None");

    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "WindowFlags_NoTitleBar");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "WindowFlags_NoResize");

    lua_pushnumber(L, 4);
    lua_setfield(L, -2, "WindowFlags_NoMove");

    lua_pushnumber(L, 8);
    lua_setfield(L, -2, "WindowFlags_NoScrollbar");

    lua_pushnumber(L, 16);
    lua_setfield(L, -2, "WindowFlags_NoScrollWithMouse");

    lua_pushnumber(L, 32);
    lua_setfield(L, -2, "WindowFlags_NoCollapse");

    lua_pushnumber(L, 43);
    lua_setfield(L, -2, "WindowFlags_NoDecoration");

    lua_pushnumber(L, 64);
    lua_setfield(L, -2, "WindowFlags_AlwaysAutoResize");

    lua_pushnumber(L, 128);
    lua_setfield(L, -2, "WindowFlags_NoBackground");

    lua_pushnumber(L, 256);
    lua_setfield(L, -2, "WindowFlags_NoSavedSettings");

    lua_pushnumber(L, 512);
    lua_setfield(L, -2, "WindowFlags_NoMouseInputs");

    lua_pushnumber(L, 1024);
    lua_setfield(L, -2, "WindowFlags_MenuBar");

    lua_pushnumber(L, 2048);
    lua_setfield(L, -2, "WindowFlags_HorizontalScrollbar");

    lua_pushnumber(L, 4096);
    lua_setfield(L, -2, "WindowFlags_NoFocusOnAppearing");

    lua_pushnumber(L, 8192);
    lua_setfield(L, -2, "WindowFlags_NoBringToFrontOnFocus");

    lua_pushnumber(L, 16384);
    lua_setfield(L, -2, "WindowFlags_AlwaysVerticalScrollbar");

    lua_pushnumber(L, 32768);
    lua_setfield(L, -2, "WindowFlags_AlwaysHorizontalScrollbar");

    lua_pushnumber(L, 65536);
    lua_setfield(L, -2, "WindowFlags_AlwaysUseWindowPadding");

    lua_pushnumber(L, 786432);
    lua_setfield(L, -2, "WindowFlags_NoNav");

    lua_pushnumber(L, 786944);
    lua_setfield(L, -2, "WindowFlags_NoInputs");

    lua_pushnumber(L, 262144);
    lua_setfield(L, -2, "WindowFlags_NoNavInputs");

    lua_pushnumber(L, 524288);
    lua_setfield(L, -2, "WindowFlags_NoNavFocus");

    lua_pushnumber(L, 1048576);
    lua_setfield(L, -2, "WindowFlags_UnsavedDocument");
}

void bindInputTextFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "InputTextFlags_None");

    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "InputTextFlags_CharsDecimal");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "InputTextFlags_CharsHexadecimal");

    lua_pushnumber(L, 4);
    lua_setfield(L, -2, "InputTextFlags_CharsUppercase");

    lua_pushnumber(L, 8);
    lua_setfield(L, -2, "InputTextFlags_CharsNoBlank");

    lua_pushnumber(L, 16);
    lua_setfield(L, -2, "InputTextFlags_AutoSelectAll");

    lua_pushnumber(L, 32);
    lua_setfield(L, -2, "InputTextFlags_EnterReturnsTrue");

    lua_pushnumber(L, 64);
    lua_setfield(L, -2, "InputTextFlags_CallbackCompletion");

    lua_pushnumber(L, 128);
    lua_setfield(L, -2, "InputTextFlags_CallbackHistory");

    lua_pushnumber(L, 256);
    lua_setfield(L, -2, "InputTextFlags_CallbackAlways");

    lua_pushnumber(L, 512);
    lua_setfield(L, -2, "InputTextFlags_CallbackCharFilter");

    lua_pushnumber(L, 1024);
    lua_setfield(L, -2, "InputTextFlags_AllowTabInput");

    lua_pushnumber(L, 2048);
    lua_setfield(L, -2, "InputTextFlags_CtrlEnterForNewLine");

    lua_pushnumber(L, 4096);
    lua_setfield(L, -2, "InputTextFlags_NoHorizontalScroll");

    lua_pushnumber(L, 8192);
    lua_setfield(L, -2, "InputTextFlags_AlwaysInsertMode");

    lua_pushnumber(L, 16384);
    lua_setfield(L, -2, "InputTextFlags_ReadOnly");

    lua_pushnumber(L, 32768);
    lua_setfield(L, -2, "InputTextFlags_Password");

    lua_pushnumber(L, 65536);
    lua_setfield(L, -2, "InputTextFlags_NoUndoRedo");

    lua_pushnumber(L, 131072);
    lua_setfield(L, -2, "InputTextFlags_CharsScientific");

    lua_pushnumber(L, 262144);
    lua_setfield(L, -2, "InputTextFlags_CallbackResize");
}

void bindTreeNodeFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "TreeNodeFlags_None");

    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "TreeNodeFlags_Selected");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "TreeNodeFlags_Framed");

    lua_pushnumber(L, 4);
    lua_setfield(L, -2, "TreeNodeFlags_AllowItemOverlap");

    lua_pushnumber(L, 8);
    lua_setfield(L, -2, "TreeNodeFlags_NoTreePushOnOpen");

    lua_pushnumber(L, 16);
    lua_setfield(L, -2, "TreeNodeFlags_NoAutoOpenOnLog");

    lua_pushnumber(L, 26);
    lua_setfield(L, -2, "TreeNodeFlags_CollapsingHeader");

    lua_pushnumber(L, 32);
    lua_setfield(L, -2, "TreeNodeFlags_DefaultOpen");

    lua_pushnumber(L, 64);
    lua_setfield(L, -2, "TreeNodeFlags_OpenOnDoubleClick");

    lua_pushnumber(L, 128);
    lua_setfield(L, -2, "TreeNodeFlags_OpenOnArrow");

    lua_pushnumber(L, 256);
    lua_setfield(L, -2, "TreeNodeFlags_Leaf");

    lua_pushnumber(L, 512);
    lua_setfield(L, -2, "TreeNodeFlags_Bullet");

    lua_pushnumber(L, 1024);
    lua_setfield(L, -2, "TreeNodeFlags_FramePadding");

    lua_pushnumber(L, 2048);
    lua_setfield(L, -2, "TreeNodeFlags_SpanAvailWidth");

    lua_pushnumber(L, 4096);
    lua_setfield(L, -2, "TreeNodeFlags_SpanFullWidth");

    lua_pushnumber(L, 8192);
    lua_setfield(L, -2, "TreeNodeFlags_NavLeftJumpsBackHere");
}

void bindPopupFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "PopupFlags_None");

    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "PopupFlags_MouseButtonDefault_");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "PopupFlags_MouseButtonMiddle");

    lua_pushnumber(L, 31);
    lua_setfield(L, -2, "PopupFlags_MouseButtonMask_");

    lua_pushnumber(L, 32);
    lua_setfield(L, -2, "PopupFlags_NoOpenOverExistingPopup");

    lua_pushnumber(L, 64);
    lua_setfield(L, -2, "PopupFlags_NoOpenOverItems");

    lua_pushnumber(L, 128);
    lua_setfield(L, -2, "PopupFlags_AnyPopupId");

    lua_pushnumber(L, 256);
    lua_setfield(L, -2, "PopupFlags_AnyPopupLevel");

    lua_pushnumber(L, 384);
    lua_setfield(L, -2, "PopupFlags_AnyPopup");
}

void bindSelectableFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "SelectableFlags_DontClosePopups");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "SelectableFlags_SpanAllColumns");

    lua_pushnumber(L, 4);
    lua_setfield(L, -2, "SelectableFlags_AllowDoubleClick");

    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "SelectableFlags_None");

    lua_pushnumber(L, 16);
    lua_setfield(L, -2, "SelectableFlags_AllowItemOverlap");

    lua_pushnumber(L, 8);
    lua_setfield(L, -2, "SelectableFlags_Disabled");
}

void bindComboFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "ComboFlags_PopupAlignLeft");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "ComboFlags_HeightSmall");

    lua_pushnumber(L, 64);
    lua_setfield(L, -2, "ComboFlags_NoPreview");

    lua_pushnumber(L, 8);
    lua_setfield(L, -2, "ComboFlags_HeightLarge");

    lua_pushnumber(L, 32);
    lua_setfield(L, -2, "ComboFlags_NoArrowButton");

    lua_pushnumber(L, 30);
    lua_setfield(L, -2, "ComboFlags_HeightMask_");

    lua_pushnumber(L, 4);
    lua_setfield(L, -2, "ComboFlags_HeightRegular");

    lua_pushnumber(L, 16);
    lua_setfield(L, -2, "ComboFlags_HeightLargest");

    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "ComboFlags_None");
}

void bindTabBarFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "TabBarFlags_Reorderable");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "TabBarFlags_AutoSelectNewTabs");

    lua_pushnumber(L, 4);
    lua_setfield(L, -2, "TabBarFlags_TabListPopupButton");

    lua_pushnumber(L, 64);
    lua_setfield(L, -2, "TabBarFlags_FittingPolicyResizeDown");

    lua_pushnumber(L, 8);
    lua_setfield(L, -2, "TabBarFlags_NoCloseWithMiddleMouseButton");

    lua_pushnumber(L, 32);
    lua_setfield(L, -2, "TabBarFlags_NoTooltip");

    lua_pushnumber(L, 128);
    lua_setfield(L, -2, "TabBarFlags_FittingPolicyScroll");

    lua_pushnumber(L, 16);
    lua_setfield(L, -2, "TabBarFlags_NoTabListScrollingButtons");

    lua_pushnumber(L, 192);
    lua_setfield(L, -2, "TabBarFlags_FittingPolicyMask_");

    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "TabBarFlags_None");

}

void bindTabItemFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "TabItemFlags_None");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "TabItemFlags_SetSelected");

    lua_pushnumber(L, 8);
    lua_setfield(L, -2, "TabItemFlags_NoPushId");

    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "TabItemFlags_UnsavedDocument");

    lua_pushnumber(L, 4);
    lua_setfield(L, -2, "TabItemFlags_NoCloseWithMiddleMouseButton");

    lua_pushnumber(L, 16);
    lua_setfield(L, -2, "TabItemFlags_NoTooltip");

}

void bindFocusedFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "FocusedFlags_ChildWindows");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "FocusedFlags_RootWindow");

    lua_pushnumber(L, 3);
    lua_setfield(L, -2, "FocusedFlags_RootAndChildWindows");

    lua_pushnumber(L, 4);
    lua_setfield(L, -2, "FocusedFlags_AnyWindow");

    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "FocusedFlags_None");

}

void bindHoveredFlagsEnum(lua_State *L)
{
    lua_pushnumber(L, 1);
    lua_setfield(L, -2, "HoveredFlags_ChildWindows");

    lua_pushnumber(L, 2);
    lua_setfield(L, -2, "HoveredFlags_RootWindow");

    lua_pushnumber(L, 3);
    lua_setfield(L, -2, "HoveredFlags_RootAndChildWindows");

    lua_pushnumber(L, 4);
    lua_setfield(L, -2, "HoveredFlags_AnyWindow");

    lua_pushnumber(L, 8);
    lua_setfield(L, -2, "HoveredFlags_AllowWhenBlockedByPopup");

    lua_pushnumber(L, 0);
    lua_setfield(L, -2, "HoveredFlags_None");

    lua_pushnumber(L, 32);
    lua_setfield(L, -2, "HoveredFlags_AllowWhenBlockedByActiveItem");

    lua_pushnumber(L, 128);
    lua_setfield(L, -2, "HoveredFlags_AllowWhenDisabled");

    lua_pushnumber(L, 104);
    lua_setfield(L, -2, "HoveredFlags_RectOnly");

    lua_pushnumber(L, 64);
    lua_setfield(L, -2, "HoveredFlags_AllowWhenOverlapped");

}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
static lua_State *L;
static Application* application;
static char keyWeak = ' ';
struct _Color {
    uint8_t r,g,b,a;
};

class GImGui
{
public:
    GImGui(LuaApplication* application, lua_State *L);
    ~GImGui();

    SpriteProxy *proxy;

    void doDraw(const CurrentTransform&, float sx, float sy, float ex, float ey);
private:
    LuaApplication* application;
    //lua_State *L;

    VertexBuffer<Point2f> vertices;
    VertexBuffer<Point2f> texcoords;
    VertexBuffer<_Color> colors;
};

static void _Draw(void *c, const CurrentTransform&t, float sx, float sy,
        float ex, float ey) {
    ((GImGui *) c)->doDraw(t, sx, sy, ex, ey);
}

static void _Destroy(void *c) {
    delete ((GImGui *) c);
}

GImGui::GImGui(LuaApplication* application, lua_State *L)
{
    this->application = application;
    proxy = gtexture_get_spritefactory()->createProxy(application->getApplication(), this, _Draw, _Destroy);
}

GImGui::~GImGui()
{

    ImGuiIO& io = ImGui::GetIO();
    io.Fonts->TexID = 0;
    ImGui::DestroyContext();
}

void GImGui::doDraw(const CurrentTransform&, float _UNUSED(sx), float _UNUSED(sy), float _UNUSED(ex), float _UNUSED(ey))
{
    ImDrawData* draw_data = ImGui::GetDrawData();
    if (!draw_data) return;

    ShaderEngine *engine=gtexture_get_engine();
    ShaderProgram *shp=engine->getDefault(ShaderEngine::STDP_TEXTURECOLOR);
    ImVec2 pos = draw_data->DisplayPos;

    for (int n = 0; n < draw_data->CmdListsCount; n++)
    {
       const ImDrawList* cmd_list = draw_data->CmdLists[n];
       const ImDrawVert* vtx_buffer = cmd_list->VtxBuffer.Data;  // vertex buffer generated by Dear ImGui
       const ImDrawIdx* idx_buffer = cmd_list->IdxBuffer.Data;   // index buffer generated by Dear ImGui
       size_t vtx_size=cmd_list->VtxBuffer.Size;
       vertices.resize(vtx_size);
       texcoords.resize(vtx_size);
       colors.resize(vtx_size);
       for (size_t i=0;i<vtx_size;i++) {
           vertices[i].x=vtx_buffer[i].pos.x;
           vertices[i].y=vtx_buffer[i].pos.y;
           texcoords[i].x=vtx_buffer[i].uv.x;
           texcoords[i].y=vtx_buffer[i].uv.y;

           uint32_t c=vtx_buffer[i].col;
           uint32_t r=(c&0xFF) >> 0;
           uint32_t g=(c&0xFF00) >> 8;
           uint32_t b=(c&0xFF0000) >> 16;
           uint32_t a=(c&0xFF000000) >> 24;

//           uint32_t r=(c >> 24);
//           uint32_t g=(c >> 16)&0xFF;
//           uint32_t b=(c >> 8)&0xFF;
//           uint32_t a=c & 0xFF;

           colors[i].r = (r * a) >> 8;
           colors[i].g = (g * a) >> 8;
           colors[i].b = (b * a) >> 8;
           colors[i].a = a;

           //colors[i]=vtx_buffer[i].col;
       }
       vertices.Update();
       texcoords.Update();
       colors.Update();
       shp->setData(ShaderProgram::DataVertex, ShaderProgram::DFLOAT,2, &vertices[0], vtx_size, true, NULL);
       shp->setData(ShaderProgram::DataTexture, ShaderProgram::DFLOAT,2, &texcoords[0], vtx_size, true, NULL);
       shp->setData(ShaderProgram::DataColor, ShaderProgram::DUBYTE,4, &colors[0], vtx_size, true, NULL);

       for (int cmd_i = 0; cmd_i < cmd_list->CmdBuffer.Size; cmd_i++)
       {
          const ImDrawCmd* pcmd = &cmd_list->CmdBuffer[cmd_i];
          if (pcmd->UserCallback)
          {
              pcmd->UserCallback(cmd_list, pcmd);
          }
          else
          {
              g_id textureId = (g_id)pcmd->TextureId;

              engine->bindTexture(0, gtexture_getInternalTexture(textureId));
              engine->pushClip((int)(pcmd->ClipRect.x - pos.x), (int)(pcmd->ClipRect.y - pos.y), (int)(pcmd->ClipRect.z - pcmd->ClipRect.x), (int)(pcmd->ClipRect.w - pcmd->ClipRect.y));
              shp->drawElements(ShaderProgram::Triangles, pcmd->ElemCount,ShaderProgram::DUSHORT, idx_buffer, true, NULL);
              engine->popClip();
         }
         idx_buffer += pcmd->ElemCount;
       }
    }
}

/// FONTS

G_FILE* GFileOpen(const char* filename, const char* mode)
{
    return g_fopen(filename, mode);
}

bool GFileClose(G_FILE* f)
{
    return g_fclose(f) == 0;
}

size_t GFileGetSize(G_FILE* f)
{
    long off = 0;
    long sz = 0;
    return ((off = g_ftell(f)) != -1 && !g_fseek(f, 0, SEEK_END) && (sz = g_ftell(f)) != -1 && !g_fseek(f, off, SEEK_SET)) ? (size_t)sz : (size_t)-1;
}

size_t GFileRead(void* data, ImU64 sz, ImU64 count, G_FILE* f)
{
    return g_fread(data, (size_t)sz, (size_t)count, f);
}

size_t GFileWrite(const void* data, ImU64 sz, ImU64 count, G_FILE* f)
{
    return g_fwrite(data, (size_t)sz, (size_t)count, f);
}

void* GFileLoadToMemory(const char* filename, const char* mode, size_t* out_file_size, int padding_bytes)
{
    IM_ASSERT(filename && mode);
    if (out_file_size)
        *out_file_size = 0;

    G_FILE *f;
    if ((f = GFileOpen(filename, mode)) == NULL)
    {
        return NULL;
    }

    size_t file_size = (size_t)GFileGetSize(f);
    if (file_size == (size_t)-1)
    {
        GFileClose(f);
        return NULL;
    }

    void* file_data = IM_ALLOC(file_size + padding_bytes);
    if (file_data == NULL)
    {
        GFileClose(f);
        return NULL;
    }
    if (GFileRead(file_data, 1, file_size, f) != file_size)
    {
        GFileClose(f);
        IM_FREE(file_data);
        return NULL;
    }
    if (padding_bytes > 0)
        memset((void*)(((char*)file_data) + file_size), 0, (size_t)padding_bytes);

    GFileClose(f);
    if (out_file_size)
        *out_file_size = file_size;

    return file_data;
}

void GAddFontFromFileTTF(const char* filename, float size_pixels, int oversample_x = 3, int oversample_y = 1, float esx = 0.0f, float esy = 0.0f)
{   ImGuiIO& io = ImGui::GetIO();

    size_t data_size = 0;
    void* data = GFileLoadToMemory(filename, "rb", &data_size, 0);
    //if (!data) luaError(L, "Cant load font file!");

    ImFontConfig config;
    config.OversampleH = oversample_x;
    config.OversampleV = oversample_y;
    config.GlyphExtraSpacing.x = esx;
    config.GlyphExtraSpacing.y = esy;

    io.Fonts->AddFontFromMemoryTTF(data, data_size, size_pixels, &config, NULL);
}

int initImGui(lua_State *L)
{
    LuaApplication* application = static_cast<LuaApplication*>(luaL_getdata(L));
    ::application = application->getApplication();

    Binder binder(L);

    int swidth = luaL_checkinteger(L, 1);
    int sheight = luaL_checkinteger(L, 2);

    GImGui *imgui = new GImGui(application, L);
    binder.pushInstance("ImGui", imgui->proxy);

    luaL_rawgetptr(L, LUA_REGISTRYINDEX, &keyWeak);
    lua_pushvalue(L, -2);
    luaL_rawsetptr(L, -2, imgui);
    lua_pop(L, 1);

    // init ImGui itself
    ImGui::CreateContext();

    // Setup style theme
    ImGui::StyleColorsDark();

    // Create font
    ImGuiIO& io = ImGui::GetIO();
    unsigned char* pixels;
    int width, height;

    io.Fonts->AddFontDefault();

    // Load fonts
    if (lua_istable(L, 3))
    {
        lua_pushvalue(L, 3);
        lua_pushnil(L);
        while (lua_next(L, -2))
        {
            lua_pushvalue(L, -2); // { fontPath, fontSize, OversampleH, OversampleV, GlyphExtraSpacingX, GlyphExtraSpacingY }

            lua_rawgeti(L, -2, 1);
            const char *filename = luaL_checkstring(L, -1);
            lua_pop(L, 1);

            lua_rawgeti(L, -2, 2);
            float size_pixel = luaL_checknumber(L, -1);
            lua_pop(L, 1);

            int oversampleH = 3;
            int oversampleV = 1;
            float spacingX = 0.0f;
            float spacingY = 0.0f;

            lua_rawgeti(L, -2, 3);
            if (!lua_isnil(L, -1)) oversampleH = luaL_checkinteger(L, -1);
            lua_pop(L, 1);

            lua_rawgeti(L, -2, 4);
            if (!lua_isnil(L, -1)) oversampleV = luaL_checkinteger(L, -1);
            lua_pop(L, 1);

            lua_rawgeti(L, -2, 5);
            if (!lua_isnil(L, -1)) spacingX = luaL_checknumber(L, -1);
            lua_pop(L, 1);

            lua_rawgeti(L, -2, 6);
            if (!lua_isnil(L, -1)) spacingY = luaL_checknumber(L, -1);
            lua_pop(L, 1);

            GAddFontFromFileTTF(filename, size_pixel, oversampleH, oversampleV, spacingX, spacingY);
            lua_pop(L, 2);
        }
        lua_pop(L, 1);
    }

    io.Fonts->GetTexDataAsRGBA32(&pixels, &width, &height);

    // Keyboard map
    // Keyboard mapping. ImGui will use those indices to peek into the io.KeyDown[] array.
    io.KeyMap[ImGuiKey_Tab] = GINPUT_KEY_TAB;
    io.KeyMap[ImGuiKey_LeftArrow] = GINPUT_KEY_LEFT;
    io.KeyMap[ImGuiKey_RightArrow] = GINPUT_KEY_RIGHT;
    io.KeyMap[ImGuiKey_UpArrow] = GINPUT_KEY_UP;
    io.KeyMap[ImGuiKey_DownArrow] = GINPUT_KEY_DOWN;
    io.KeyMap[ImGuiKey_PageUp] = GINPUT_KEY_PAGEUP;
    io.KeyMap[ImGuiKey_PageDown] = GINPUT_KEY_PAGEDOWN;
    io.KeyMap[ImGuiKey_Home] = GINPUT_KEY_HOME;
    io.KeyMap[ImGuiKey_End] = GINPUT_KEY_END;
    io.KeyMap[ImGuiKey_Delete] = GINPUT_KEY_DELETE;
    io.KeyMap[ImGuiKey_Backspace] = GINPUT_KEY_BACKSPACE;
    io.KeyMap[ImGuiKey_Enter] = GINPUT_KEY_ENTER;
    io.KeyMap[ImGuiKey_Escape] = GINPUT_KEY_ESC;
    io.KeyMap[ImGuiKey_Insert] = GINPUT_KEY_INSERT;
    io.KeyMap[ImGuiKey_A] = GINPUT_KEY_A;
    io.KeyMap[ImGuiKey_C] = GINPUT_KEY_C;
    io.KeyMap[ImGuiKey_V] = GINPUT_KEY_V;
    io.KeyMap[ImGuiKey_X] = GINPUT_KEY_X;
    io.KeyMap[ImGuiKey_Y] = GINPUT_KEY_Y;
    io.KeyMap[ImGuiKey_Z] = GINPUT_KEY_Z;

    g_id texture = gtexture_create(width, height, GTEXTURE_RGBA, GTEXTURE_UNSIGNED_BYTE, GTEXTURE_CLAMP, GTEXTURE_NEAREST, pixels, NULL, 0);
    io.Fonts->TexID = (void *)texture;

    io.DisplaySize.x = swidth;
    io.DisplaySize.y = sheight;


    return 1;
}

int destroyImGui(lua_State *L)
{
    ImGuiIO& io = ImGui::GetIO();

    io.Fonts->TexID = 0;

    ImGui::DestroyContext();

    return 0;
}

////////////////////////////////////////////////////////////////////////////////
///
/// BINDINGS.
///
////////////////////////////////////////////////////////////////////////////////

/// MOUSE INPUTS

int ImGui_impl_MouseHover(lua_State *L)
{
    ImGuiIO& io = ImGui::GetIO();
    io.MousePos.x = getfield(L, "x");
    io.MousePos.y = getfield(L, "y");

    return 0;
}

int ImGui_impl_MouseMove(lua_State *L)
{
    int button = (int)getfield(L, "button");

    // shift 1 bit to translate gideros buttons: 1, 2, 4
    // to ImGui buttons: 0, 1, 2
    button >>= 1;

    ImGuiIO& io = ImGui::GetIO();
    io.MousePos.x = getfield(L, "x");
    io.MousePos.y = getfield(L, "y");
    io.MouseDown[button] = true;

    return 0;
}

int ImGui_impl_MouseDown(lua_State *L)
{
    int button = (int)getfield(L, "button");

    // shift 1 bit to translate gideros buttons: 1, 2, 4
    // to ImGui buttons: 0, 1, 2
    button >>= 1;

    ImGuiIO& io = ImGui::GetIO();
    io.MousePos.x = getfield(L, "x");
    io.MousePos.y = getfield(L, "y");
    io.MouseDown[button] = true;

    return 0;
}

int ImGui_impl_MouseUp(lua_State *L)
{
    int button = (int)getfield(L, "button");

    // shift 1 bit to translate gideros buttons: 1, 2, 4
    // to ImGui buttons: 0, 1, 2
    button >>= 1;

    ImGuiIO& io = ImGui::GetIO();
    io.MousePos.x = getfield(L, "x");
    io.MousePos.y = getfield(L, "y");
    io.MouseDown[button] = false;

    return 0;
}

int ImGui_impl_MouseWheel(lua_State *L)
{
    float wheel = getfield(L, "wheel");

    ImGuiIO& io = ImGui::GetIO();
    io.MouseWheel += wheel < 0 ? -1.0f : 1.0f;
    io.MousePos.x = getfield(L, "x");
    io.MousePos.y = getfield(L, "y");

    return 0;
}

/// KEYBOARD INPUTS

int ImGui_impl_KeyUp(lua_State *L)
{
    int keyCode = (int)getfield(L, "keyCode");

    ImGuiIO& io = ImGui::GetIO();
    io.KeysDown[keyCode] = false;

    int mod = getKeyboardModifiers(L);
    io.KeyAlt = (mod & GINPUT_ALT_MODIFIER) > 0;
    io.KeyCtrl = (mod & GINPUT_CTRL_MODIFIER) > 0;
    io.KeyShift = (mod & GINPUT_SHIFT_MODIFIER) > 0;
    io.KeySuper = (mod & GINPUT_META_MODIFIER) > 0;

    return 0;
}

int ImGui_impl_KeyDown(lua_State *L)
{
    int keyCode = (int)getfield(L, "keyCode");

    ImGuiIO& io = ImGui::GetIO();
    io.KeysDown[keyCode] = true;

    int mod = getKeyboardModifiers(L);
    io.KeyAlt = (mod & GINPUT_ALT_MODIFIER) > 0;
    io.KeyCtrl = (mod & GINPUT_CTRL_MODIFIER) > 0;
    io.KeyShift = (mod & GINPUT_SHIFT_MODIFIER) > 0;
    io.KeySuper = (mod & GINPUT_META_MODIFIER) > 0;

    return 0;
}

int ImGui_impl_KeyChar(lua_State *L)
{
    lua_pushstring(L, "text");
    lua_gettable(L, -2);
    const char *text = lua_tostring(L, -1);
    lua_pop(L, 1);

    ImGuiIO& io = ImGui::GetIO();
    io.AddInputCharactersUTF8(text);

    return 0;
}

/// INPUT FLAGS

int ImGui_impl_WantCaptureMouse(lua_State *L)
{
    ImGuiIO& io = ImGui::GetIO();

    lua_pushboolean(L, io.WantCaptureMouse);
    return 1;
}

int ImGui_impl_WantCaptureKeyboard(lua_State *L)
{
    ImGuiIO& io = ImGui::GetIO();

    lua_pushboolean(L, io.WantCaptureKeyboard);
    return 1;
}

int ImGui_impl_WantTextInput(lua_State *L)
{
    ImGuiIO& io = ImGui::GetIO();

    lua_pushboolean(L, io.WantTextInput);
    return 1;
}

/// DRAWING STUFF

int ImGui_impl_NewFrame(lua_State *L)
{
    ::L = L;

    double deltaTime = getfield(L, "deltaTime");

    ImGuiIO& io = ImGui::GetIO();
    io.DeltaTime = deltaTime;

    ImGui::NewFrame();
    //ImGui::ShowMetricsWindow();

    return 0;
}

int ImGui_impl_Render(lua_State *L)
{
    ImGui::Render();
    return 0;
}

int ImGui_impl_EndFrame(lua_State *L)
{
    ImGui::EndFrame();
    return 0;
}

// Windows
int ImGui_impl_Begin(lua_State *L)
{
    const char *name = luaL_checkstring(L, 2);
    bool p_open = lua_toboolean(L, 3) > 0;
    int flags = luaL_optinteger(L, 4, 0);

    bool result = ImGui::Begin(name, &p_open, flags);

    lua_pushboolean(L, result);
    lua_pushboolean(L, p_open);

    return 2;
}

int ImGui_impl_End(lua_State *L)
{
    ImGui::End();
    return 0;
}

// Child Windows
int ImGui_impl_BeginChild(lua_State *L)
{
    float w = luaL_optnumber(L, 3, 0);
    float h = luaL_optnumber(L, 4, 0);
    bool border = lua_toboolean(L, 5) > 0;
    int flags = luaL_optinteger(L, 6, 0);
    bool result;

    if (lua_type(L, 2) == LUA_TSTRING)
    {
        const char *str_id = luaL_checkstring(L, 2);
        result = ImGui::BeginChild(str_id, ImVec2(w, h), border, flags);
    }
    else
    {
        ImGuiID id = luaL_checkinteger(L, 2);
        result = ImGui::BeginChild(id, ImVec2(w, h), border, flags);
    }

    lua_pushboolean(L, result);

    return 1;
}

int ImGui_impl_EndChild(lua_State *L)
{
    ImGui::EndChild();
    return 0;
}

// Windows Utilities
int ImGui_impl_IsWindowAppearing(lua_State *L)
{
    lua_pushboolean(L, ImGui::IsWindowAppearing());
    return 1;
}

int ImGui_impl_IsWindowCollapsed(lua_State *L)
{
    lua_pushboolean(L, ImGui::IsWindowCollapsed());
    return 1;
}

int ImGui_impl_IsWindowFocused(lua_State *L)
{
    int flag = luaL_optinteger(L, 2, 0);
    lua_pushboolean(L, ImGui::IsWindowFocused(flag));
    return 1;
}

int ImGui_impl_IsWindowHovered(lua_State *L)
{
    int flag = luaL_optinteger(L, 2, 0);
    lua_pushboolean(L, ImGui::IsWindowHovered(flag));
    return 1;
}

int ImGui_impl_GetWindowPos(lua_State *L)
{
    ImVec2 pos = ImGui::GetWindowPos();
    lua_pushnumber(L, pos.x);
    lua_pushnumber(L, pos.y);
    return  2;
}

int ImGui_impl_GetWindowSize(lua_State *L)
{
    ImVec2 size = ImGui::GetWindowSize();
    lua_pushnumber(L, size.x);
    lua_pushnumber(L, size.y);
    return  2;
}

int ImGui_impl_GetWindowWidth(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetWindowWidth());
    return  1;
}

int ImGui_impl_GetWindowHeight(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetWindowHeight());
    return  1;
}

int ImGui_impl_SetNextWindowPos(lua_State *L)
{
    float x = luaL_checknumber(L, 2);
    float y = luaL_checknumber(L, 3);
    const ImVec2& pos = ImVec2(x, y);
    ImGuiCond cond = luaL_optinteger(L, 4, 0);
    float px = luaL_optnumber(L, 5, 0);
    float py = luaL_optnumber(L, 6, 0);
    const ImVec2& pivot = ImVec2(px, py);

    ImGui::SetNextWindowPos(pos, cond, pivot);

    return 0;
}

int ImGui_impl_SetNextWindowSize(lua_State *L)
{
    float w = luaL_checknumber(L, 2);
    float h = luaL_checknumber(L, 3);
    const ImVec2& size = ImVec2(w, h);
    ImGuiCond cond = luaL_optinteger(L, 4, 0);

    ImGui::SetNextWindowSize(size, cond);

    return 0;
}

//int ImGui_impl_SetNextWindowSizeConstraints(lua_State *L)
//{
//    const ImVec2& size_min, const ImVec2& size_max, ImGuiSizeCallback custom_callback = NULL, void* custom_callback_data = NULL
//    return 0;
//}

int ImGui_impl_SetNextWindowContentSize(lua_State *L)
{
    float w = luaL_checknumber(L, 2);
    float h = luaL_checknumber(L, 3);
    const ImVec2& size = ImVec2(w, h);

    ImGui::SetNextWindowContentSize(size);

    return 0;
}

int ImGui_impl_SetNextWindowCollapsed(lua_State *L)
{
    // bool collapsed, ImGuiCond cond = 0
    bool collapsed = lua_toboolean(L, 2) > 0;
    ImGuiCond cond = luaL_optinteger(L, 3, 0);

    ImGui::SetNextWindowCollapsed(collapsed, cond);
    return 0;
}

int ImGui_impl_SetNextWindowFocus(lua_State *L)
{
    ImGui::SetNextWindowFocus();
    return 0;
}

int ImGui_impl_SetNextWindowBgAlpha(lua_State *L)
{
    float alpha = luaL_checknumber(L, 2);
    ImGui::SetNextWindowBgAlpha(alpha);
    return 0;
}

int ImGui_impl_SetWindowPos(lua_State *L)
{
    if (lua_type(L, 2) == LUA_TSTRING)
    {
        const char *name = luaL_checkstring(L, 2);
        float x = luaL_checknumber(L, 3);
        float y = luaL_checknumber(L, 4);
        const ImVec2& pos = ImVec2(x, y);
        ImGuiCond cond = luaL_optinteger(L, 5, 0);

        ImGui::SetWindowPos(name, pos, cond);
    }
    else
    {
        float x = luaL_checknumber(L, 2);
        float y = luaL_checknumber(L, 3);
        const ImVec2& pos = ImVec2(x, y);
        ImGuiCond cond = luaL_optinteger(L, 4, 0);

        ImGui::SetWindowPos(pos, cond);
    }

    return  0;
}

int ImGui_impl_SetWindowSize(lua_State *L)
{
    if (lua_type(L, 2) == LUA_TSTRING)
    {
        const char *name = luaL_checkstring(L, 2);
        float w = luaL_checknumber(L, 3);
        float h = luaL_checknumber(L, 4);
        const ImVec2& size = ImVec2(w, h);
        ImGuiCond cond = luaL_optinteger(L, 5, 0);

        ImGui::SetWindowSize(name, size, cond);
    }
    else
    {
        float w = luaL_checknumber(L, 2);
        float h = luaL_checknumber(L, 3);
        const ImVec2& size = ImVec2(w, h);
        ImGuiCond cond = luaL_optinteger(L, 4, 0);

        ImGui::SetWindowSize(size, cond);
    }

    return 0;
}

int ImGui_impl_SetWindowCollapsed(lua_State *L)
{
    if (lua_type(L, 2) == LUA_TSTRING)
    {
        const char *name = luaL_checkstring(L, 2);
        bool collapsed = lua_toboolean(L, 3) > 0;
        ImGuiCond cond = luaL_optinteger(L, 4, 0);

        ImGui::SetWindowCollapsed(name, collapsed, cond);
    }
    else
    {
        bool collapsed = lua_toboolean(L, 2) > 0;
        ImGuiCond cond = luaL_optinteger(L, 3, 0);

        ImGui::SetWindowCollapsed(collapsed, cond);
    }

    return 0;
}

int ImGui_impl_SetWindowFocus(lua_State *L)
{
    if (lua_type(L, 2) == LUA_TSTRING)
    {
        const char *name = luaL_checkstring(L, 2);
        ImGui::SetWindowFocus(name);
    }
    else
        ImGui::SetWindowFocus();

    return 0;
}

//int ImGui_impl_SetWindowFontScale(lua_State *L)
//{
//    float scale = luaL_checknumber(L, 2);
//    ImGui::SetWindowFontScale(scale);
//    return 0;
//}

// Content region
// - Those functions are bound to be redesigned soon (they are confusing, incomplete and return values in local window coordinates which increases confusion)
int ImGui_impl_GetContentRegionMax(lua_State *L)
{
    ImVec2 max = ImGui::GetContentRegionMax();
    lua_pushnumber(L, max.x);
    lua_pushnumber(L, max.y);
    return 2;
}

int ImGui_impl_GetContentRegionAvail(lua_State *L)
{
    ImVec2 avail = ImGui::GetContentRegionAvail();
    lua_pushnumber(L, avail.x);
    lua_pushnumber(L, avail.y);
    return 2;
}

int ImGui_impl_GetWindowContentRegionMin(lua_State *L)
{
    ImVec2 min = ImGui::GetWindowContentRegionMin();
    lua_pushnumber(L, min.x);
    lua_pushnumber(L, min.y);
    return 2;
}

int ImGui_impl_GetWindowContentRegionMax(lua_State *L)
{
    ImVec2 max = ImGui::GetWindowContentRegionMax();
    lua_pushnumber(L, max.x);
    lua_pushnumber(L, max.y);
    return 2;
}

int ImGui_impl_GetWindowContentRegionWidth(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetWindowContentRegionWidth());
    return 1;
}

// Windows Scrolling
int ImGui_impl_GetScrollX(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetScrollX());
    return 1;
}

int ImGui_impl_GetScrollY(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetScrollY());
    return 1;
}

int ImGui_impl_GetScrollMaxX(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetScrollMaxX());
    return 1;
}

int ImGui_impl_GetScrollMaxY(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetScrollMaxY());
    return 1;
}

int ImGui_impl_SetScrollX(lua_State *L)
{
    float scroll_x = luaL_checknumber(L, 2);
    ImGui::SetScrollX(scroll_x);
    return 0;
}

int ImGui_impl_SetScrollY(lua_State *L)
{
    float scroll_y = luaL_checknumber(L, 2);
    ImGui::SetScrollY(scroll_y);
    return 0;
}

int ImGui_impl_SetScrollHereX(lua_State *L)
{
    float center_x_ratio = luaL_optnumber(L, 2, 0.5f);
    ImGui::SetScrollHereX(center_x_ratio);
    return 0;
}

int ImGui_impl_SetScrollHereY(lua_State *L)
{
    float center_y_ratio = luaL_optnumber(L, 2, 0.5f);
    ImGui::SetScrollHereY(center_y_ratio);
    return 0;
}

int ImGui_impl_SetScrollFromPosX(lua_State *L)
{
    float local_x = luaL_checknumber(L, 2);
    float center_x_ratio = luaL_optnumber(L, 3, 0.5f);
    ImGui::SetScrollFromPosX(local_x, center_x_ratio);
    return 0;
}

int ImGui_impl_SetScrollFromPosY(lua_State *L)
{
    float local_y = luaL_checknumber(L, 2);
    float center_y_ratio = luaL_optnumber(L, 3, 0.5f);
    ImGui::SetScrollFromPosY(local_y, center_y_ratio);
    return 0;
}

// Parameters stacks (shared)
int ImGui_impl_PushStyleColor(lua_State *L)
{
    ImGuiCol idx = luaL_checkinteger(L, 2);

    if (lua_type(L, 4) == LUA_TNIL)
    {
        ImU32 col = luaL_checkinteger(L, 3);
        ImGui::PushStyleColor(idx, col);
    }
    else
    {
        ImU32 r = luaL_checknumber(L, 3);
        ImU32 g = luaL_checknumber(L, 4);
        ImU32 b = luaL_checknumber(L, 5);
        ImU32 a = luaL_optnumber(L, 6, 1);
        ImGui::PushStyleColor(idx, ImVec4(r, g, b, a));
    }

    return 0;
}

int ImGui_impl_PopStyleColor(lua_State *L)
{
    int count = luaL_optinteger(L, 2, 1);
    ImGui::PopStyleColor(count);
    return 0;
}

int ImGui_impl_PushStyleVar(lua_State *L)
{
    ImGuiStyleVar idx = luaL_checkinteger(L, 2);

    if (!lua_type(L, 4) == LUA_TNIL)
    {
        float vx = luaL_checknumber(L, 3);
        float vy = luaL_checknumber(L, 4);
        ImGui::PushStyleVar(idx, ImVec2(vx, vy));
    }
    else
    {
        float val = luaL_checknumber(L, 3);
        ImGui::PushStyleVar(idx, val);
    }
    return 0;
}

int ImGui_impl_PopStyleVar(lua_State *L)
{
    int count = luaL_optinteger(L, 2, 1);
    ImGui::PopStyleVar(count);
    return 0;
}

int ImGui_impl_GetStyleColorVec4(lua_State *L)
{
    ImGuiCol idx = luaL_checkinteger(L, 2);
    const ImVec4 col = ImGui::GetStyleColorVec4(idx);
    lua_pushnumber(L, col.x);
    lua_pushnumber(L, col.y);
    lua_pushnumber(L, col.w);
    lua_pushnumber(L, col.z);
    return 4;
}

int ImGui_impl_GetFontSize(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetFontSize());
    return 1;
}

int ImGui_impl_GetColorU32(lua_State *L)
{
    if (lua_isnil(L, 3) == LUA_TNIL)
    {
        ImU32 col = luaL_checkinteger(L, 2);
        lua_pushinteger(L, ImGui::GetColorU32(col));
    }
    else if (lua_isnil(L, 4) == LUA_TNIL)
    {
        ImGuiCol idx = luaL_checkinteger(L, 2);
        float alpha_mul = luaL_optnumber(L, 3, 1.0f);
        lua_pushinteger(L, ImGui::GetColorU32(idx, alpha_mul));
    }
    else
    {
        float r = luaL_checknumber(L, 2);
        float g = luaL_checknumber(L, 3);
        float b = luaL_checknumber(L, 4);
        float a = luaL_optnumber(L, 5, 1.0f);
        lua_pushinteger(L, ImGui::GetColorU32(ImVec4(r, g, b, a)));
    }
    return 1;
}

// Parameters stacks (current window)
int ImGui_impl_PushItemWidth(lua_State *L)
{
    float item_width = luaL_checknumber(L, 2);
    ImGui::PushItemWidth(item_width);
    return 0;
}

int ImGui_impl_PopItemWidth(lua_State *L)
{
    ImGui::PopItemWidth();
    return 0;
}

int ImGui_impl_SetNextItemWidth(lua_State *L)
{
    float item_width = luaL_checknumber(L, 2);
    ImGui::SetNextItemWidth(item_width);
    return 0;
}

int ImGui_impl_CalcItemWidth(lua_State *L)
{
    lua_pushnumber(L, ImGui::CalcItemWidth());
    return 1;
}

int ImGui_impl_PushTextWrapPos(lua_State *L)
{
    float wrap_local_pos_x = luaL_optnumber(L, 2, 0.0f);
    ImGui::PushTextWrapPos(wrap_local_pos_x);
    return 0;
}

int ImGui_impl_PopTextWrapPos(lua_State *L)
{
    ImGui::PopTextWrapPos();
    return 0;
}

int ImGui_impl_PushAllowKeyboardFocus(lua_State *L)
{
    bool allow_keyboard_focus = lua_toboolean(L, 2) > 0;
    ImGui::PushAllowKeyboardFocus(allow_keyboard_focus);
    return 0;
}

int ImGui_impl_PopAllowKeyboardFocus(lua_State *L)
{
    ImGui::PopAllowKeyboardFocus();
    return 0;
}

int ImGui_impl_PushButtonRepeat(lua_State *L)
{
    bool repeat = lua_toboolean(L, 2) > 0;
    ImGui::PushButtonRepeat(repeat);
    return 0;
}

int ImGui_impl_PopButtonRepeat(lua_State *L)
{
    ImGui::PopButtonRepeat();
    return 0;
}

// Cursor / Layout
int ImGui_impl_Separator(lua_State *L)
{
    ImGui::Separator();
    return 0;
}

int ImGui_impl_SameLine(lua_State *L)
{
    float offset_from_start_x = luaL_optnumber(L, 2, 0.0f);
    float spacing = luaL_optnumber(L, 2, -1.0f);
    ImGui::SameLine(offset_from_start_x, spacing);
    return 0;
}

int ImGui_impl_NewLine(lua_State *L)
{
    ImGui::NewLine();
    return 0;
}

int ImGui_impl_Spacing(lua_State *L)
{
    ImGui::Spacing();
    return 0;
}

int ImGui_impl_Dummy(lua_State *L)
{
    float w = luaL_checknumber(L, 2);
    float h = luaL_checknumber(L, 3);

    ImGui::Dummy(ImVec2(w, h));
    return 0;
}

int ImGui_impl_Indent(lua_State *L)
{
    float indent_w = luaL_optnumber(L, 2, 0.0f);
    ImGui::Indent(indent_w);
    return 0;
}

int ImGui_impl_Unindent(lua_State *L)
{
    float indent_w = luaL_optnumber(L, 2, 0.0f);
    ImGui::Unindent(indent_w);
    return 0;
}

int ImGui_impl_BeginGroup(lua_State *L)
{
    ImGui::BeginGroup();
    return 0;
}

int ImGui_impl_EndGroup(lua_State *L)
{
    ImGui::EndGroup();
    return 0;
}

int ImGui_impl_GetCursorPos(lua_State *L)
{
    ImVec2 pos = ImGui::GetCursorPos();
    lua_pushnumber(L, pos.x);
    lua_pushnumber(L, pos.y);
    return 2;
}

int ImGui_impl_GetCursorPosX(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetCursorPosX());
    return 1;
}

int ImGui_impl_GetCursorPosY(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetCursorPosY());
    return 1;
}

//void SetCursorPos(const ImVec2& local_pos);

//void SetCursorPosX(float local_x);

//void SetCursorPosY(float local_y);

int ImGui_impl_GetCursorStartPos(lua_State *L)
{
    ImVec2 pos = ImGui::GetCursorStartPos();
    lua_pushnumber(L, pos.x);
    lua_pushnumber(L, pos.y);
    return 2;
}

int ImGui_impl_GetCursorScreenPos(lua_State *L)
{
    ImVec2 pos = ImGui::GetCursorScreenPos();
    lua_pushnumber(L, pos.x);
    lua_pushnumber(L, pos.y);
    return 2;
}

//void SetCursorScreenPos(const ImVec2& pos);

int ImGui_impl_AlignTextToFramePadding(lua_State *L)
{
    ImGui::AlignTextToFramePadding();
    return 0;
}

int ImGui_impl_GetTextLineHeight(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetTextLineHeight());
    return 1;
}

int ImGui_impl_GetTextLineHeightWithSpacing(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetTextLineHeightWithSpacing());
    return 1;
}

int ImGui_impl_GetFrameHeight(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetFrameHeight());
    return 1;
}

int ImGui_impl_GetFrameHeightWithSpacing(lua_State *L)
{
    lua_pushnumber(L, ImGui::GetFrameHeightWithSpacing());
    return 1;
}

int ImGui_impl_PushID_string(lua_State *L)
{
    if (lua_type(L, 3) == LUA_TNIL)
    {
        ImGui::PushID(luaL_checkstring(L, 2));
    }
    else
    {
        ImGui::PushID(luaL_checkstring(L, 2), luaL_checkstring(L, 3));
    }
    return 0;
}

int ImGui_impl_PushID_number(lua_State *L)
{
    ImGui::PushID(luaL_checkinteger(L, 2));
    return 0;
}

int ImGui_impl_PopID(lua_State *L)
{
    ImGui::PopID();
    return 0;
}

int ImGui_impl_GetID_string(lua_State *L)
{

    if (lua_type(L, 3) == LUA_TNIL)
    {
        lua_pushnumber(L, ImGui::GetID(luaL_checkstring(L, 2)));
    }
    else
    {
        lua_pushnumber(L, ImGui::GetID(luaL_checkstring(L, 2), luaL_checkstring(L, 3)));
    }
    return 1;
}

// Widgets: Text

int ImGui_impl_TextUnformatted(lua_State *L)
{
    const char* text = luaL_checkstring(L, 2);
    const char* text_end = luaL_optstring(L, 3, NULL);
    ImGui::TextUnformatted(text, text_end);
    return 0;
}

int ImGui_impl_Text(lua_State *L)
{
    const char* text = luaL_checkstring(L, 2);
    ImGui::Text("%s", text);
    return 0;
}

int ImGui_impl_TextColored(lua_State *L)
{
    const char* text = luaL_checkstring(L, 2);
    float r = luaL_checknumber(L, 3);
    float g = luaL_checknumber(L, 4);
    float b = luaL_checknumber(L, 5);
    float a = luaL_optnumber(L, 6, 1.0f);
    ImGui::TextColored(ImVec4(r, g, b, a), "%s", text);
    return 0;
}

int ImGui_impl_TextDisabled(lua_State *L)
{
    const char* text = luaL_checkstring(L, 2);
    ImGui::TextDisabled("%s", text);
    return 0;
}

int ImGui_impl_TextWrapped(lua_State *L)
{
    const char* text = luaL_checkstring(L, 2);
    ImGui::TextWrapped("%s", text);
    return 0;
}

int ImGui_impl_LabelText(lua_State *L)
{
    const char* text = luaL_checkstring(L, 2);
    const char* label = luaL_checkstring(L, 3);
    ImGui::LabelText(label, "%s", text);
    return 0;
}

int ImGui_impl_BulletText(lua_State *L)
{
    const char* text = luaL_checkstring(L, 2);
    ImGui::BulletText("%s", text);
    return 0;
}

// Widgets: Main
int ImGui_impl_Button(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    float w = luaL_optnumber(L, 3, 0.0f);
    float h = luaL_optnumber(L, 4, 0.0f);
    const ImVec2& size = ImVec2(w, h);
    lua_pushboolean(L, ImGui::Button(label, size));
    return 1;
}

int ImGui_impl_SmallButton(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    lua_pushboolean(L, ImGui::SmallButton(label));
    return 1;
}

int ImGui_impl_InvisibleButton(lua_State *L)
{
    const char* str_id = luaL_checkstring(L, 2);
    float w = luaL_optnumber(L, 3, 0.0f);
    float h = luaL_optnumber(L, 4, 0.0f);
    const ImVec2& size = ImVec2(w, h);
    lua_pushboolean(L, ImGui::InvisibleButton(str_id, size));
    return 1;
}

int ImGui_impl_ArrowButton(lua_State *L)
{
    const char* str_id = luaL_checkstring(L, 2);
    ImGuiDir dir = luaL_checknumber(L, 3);
    lua_pushboolean(L, ImGui::ArrowButton(str_id, dir));
    return 1;
}

int ImGui_impl_Image(lua_State *L)
{
    MyTextureData data = getTexture(L, 2);
    const ImVec2& size = ImVec2(luaL_checknumber(L, 3), luaL_checknumber(L, 4));
    ImVec4 tint = getColor(L, 5, 0xffffffff);
    ImVec4 border = getColor(L, 6);
    const ImVec2& uv1 = ImVec2(luaL_optnumber(L, 7, data.width), luaL_optnumber(L, 8, data.height));
    const ImVec2& uv0 = ImVec2(luaL_optnumber(L, 9, 0.0f), luaL_optnumber(L, 10, 0.0f));

    ImGui::Image(data.texture, size, uv0, uv1, tint, border);
    return 0;
}

int ImGui_impl_ImageButton(lua_State *L)
{
    MyTextureData data = getTexture(L, 2);
    const ImVec2& size = ImVec2(luaL_checknumber(L, 3), luaL_checknumber(L, 4));
    int frame_padding = luaL_optinteger(L, 5, -1);
    ImVec4 bg_col = getColor(L, 6);
    ImVec4 tint = getColor(L, 7, 0xffffffff);
    const ImVec2& uv1 = ImVec2(luaL_optnumber(L, 8, data.width), luaL_optnumber(L, 9, data.height));
    const ImVec2& uv0 = ImVec2(luaL_optnumber(L, 10, 0.0f), luaL_optnumber(L, 11, 0.0f));

    lua_pushboolean(L, ImGui::ImageButton(data.texture, size, uv0, uv1, frame_padding, bg_col, tint));
    return 1;
}

int ImGui_impl_Checkbox(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    bool v = lua_toboolean2(L, 3) > 0;
    lua_pushboolean(L, ImGui::Checkbox(label, &v));
    return  1;
}

// TODO
int ImGui_impl_CheckboxFlags(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    unsigned int* flags = 0;
    unsigned int flags_value = 0;

    lua_pushboolean(L, ImGui::CheckboxFlags(label, flags, flags_value));
    return 1;
}

int ImGui_impl_RadioButton(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    bool active = lua_toboolean2(L, 3) > 0;
    lua_pushboolean(L, ImGui::RadioButton(label, active));
    return 1;
}

int ImGui_impl_ProgressBar(lua_State *L)
{
    float fraction = luaL_checknumber(L, 2);
    const ImVec2& size = ImVec2(luaL_optnumber(L, 3, -1.0f), luaL_optnumber(L, 4, 0.0f));
    const char* overlay = luaL_optstring(L, 5, NULL);
    ImGui::ProgressBar(fraction, size, overlay);
    return  0;
}

int ImGui_impl_Bullet(lua_State *L)
{
    ImGui::Bullet();
    return 0;
}

// Widgets: Combo Box
int ImGui_impl_BeginCombo(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    const char* preview_value = luaL_checkstring(L, 3);
    ImGuiComboFlags flags = luaL_optinteger(L, 4, 0);
    lua_pushboolean(L, ImGui::BeginCombo(label, preview_value, flags));
    return 1;
}

int ImGui_impl_EndCombo(lua_State *L)
{
    ImGui::EndCombo();
    return 0;
}

int ImGui_impl_Combo(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    int item_current = luaL_checkinteger(L, 3);
    luaL_checktype(L, 4, LUA_TTABLE);
    size_t len = lua_objlen(L, 4);
    const char* items[len];
    lua_pushvalue(L, 4);
    lua_pushnil(L);
    while (lua_next(L, -2))
    {
        lua_pushvalue(L, -2);
        int index = lua_tointeger(L, -1);
        const char *str = lua_tostring(L, -2);
        items[index] = str;
        lua_pop(L, 2);
    }
    lua_pop(L, 1);
    lua_pushboolean(L, ImGui::Combo(label, &item_current, items, len));
    free(items);
    return 1;
}

// Widgets: Drags
// return 1
int ImGui_impl_DragFloat(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    float v = luaL_checknumber(L, 3);
    float v_speed = luaL_optnumber(L, 4, 1.0f);
    float v_min = luaL_optnumber(L, 5, 0.0f);
    float v_max = luaL_optnumber(L, 6, 0.0f);
    const char* format = luaL_optstring(L, 7, "%.3f");
    float power = luaL_optnumber(L, 8, 1.0f);

    bool result = ImGui::DragFloat(label, &v, v_speed, v_min, v_max, format, power);

    lua_pushnumber(L, v);
    lua_pushboolean(L, result);

    return 2;
}

int ImGui_impl_DragFloat2(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    ImVec2 v = ImVec2(luaL_checknumber(L, 3), luaL_checknumber(L, 4));
    float v_speed = luaL_optnumber(L, 5, 1.0f);
    float v_min = luaL_optnumber(L, 6, 0.0f);
    float v_max = luaL_optnumber(L, 7, 0.0f);
    const char* format = luaL_optstring(L, 8, "%.3f");
    float power = luaL_optnumber(L, 9, 1.0f);
    bool result = ImGui::DragFloat2(label, (float*)&v, v_speed, v_min, v_max, format, power);

    lua_pushnumber(L, v.x);
    lua_pushnumber(L, v.y);
    lua_pushboolean(L, result);
    return 3;
}

int ImGui_impl_DragFloat3(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    ImVec4 v = ImVec4(luaL_checknumber(L, 3), luaL_checknumber(L, 4), luaL_checknumber(L, 5), 0.0f);
    float v_speed = luaL_optnumber(L, 6, 1.0f);
    float v_min = luaL_optnumber(L, 7, 0.0f);
    float v_max = luaL_optnumber(L, 8, 0.0f);
    const char* format = luaL_optstring(L, 9, "%.3f");
    float power = luaL_optnumber(L, 10, 1.0f);
    bool result = ImGui::DragFloat3(label, (float*)&v, v_speed, v_min, v_max, format, power);

    lua_pushnumber(L, v.x);
    lua_pushnumber(L, v.y);
    lua_pushnumber(L, v.z);
    lua_pushboolean(L, result);
    return 4;
}

int ImGui_impl_DragFloat4(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    ImVec4 v = ImVec4(luaL_checknumber(L, 3), luaL_checknumber(L, 4), luaL_checknumber(L, 5), luaL_checknumber(L, 6));
    float v_speed = luaL_optnumber(L, 7, 1.0f);
    float v_min = luaL_optnumber(L, 8, 0.0f);
    float v_max = luaL_optnumber(L, 9, 0.0f);
    const char* format = luaL_optstring(L, 10, "%.3f");
    float power = luaL_optnumber(L, 11, 1.0f);
    bool result = ImGui::DragFloat4(label, (float*)&v, v_speed, v_min, v_max, format, power);

    lua_pushnumber(L, v.x);
    lua_pushnumber(L, v.y);
    lua_pushnumber(L, v.z);
    lua_pushnumber(L, v.w);
    lua_pushboolean(L, result);
    return 5;
}

int ImGui_impl_DragFloatRange2(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    float v_current_min = luaL_checknumber(L, 3);
    float v_current_max = luaL_checknumber(L, 4);
    float v_speed = luaL_optnumber(L, 5, 1.0f);
    float v_min = luaL_optnumber(L, 6, 0.0f);
    float v_max = luaL_optnumber(L, 7, 0.0f);
    const char* format = luaL_optstring(L, 8, "%.3f");
    const char* format_max = luaL_optstring(L, 9, NULL);
    float power = luaL_optnumber(L, 10, 1.0f);

    bool result = ImGui::DragFloatRange2(label, &v_current_min, &v_current_max, v_speed, v_min, v_max, format, format_max, power);

    lua_pushnumber(L, v_current_min);
    lua_pushnumber(L, v_current_max);
    lua_pushboolean(L, result);
    return 3;
}

int ImGui_impl_DragInt(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    int v = luaL_checkinteger(L, 3);
    float v_speed = luaL_optnumber(L, 4, 1.0f);
    int v_min = luaL_optinteger(L, 5, 0);
    int v_max = luaL_optinteger(L, 6, 0);
    const char* format = luaL_optstring(L, 7, "%d");

    bool result = ImGui::DragInt(label, &v, v_speed, v_min, v_max, format);

    lua_pushinteger(L, v);
    lua_pushboolean(L, result);
    return 2;
}

int ImGui_impl_DragInt2(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    static int vec2i[2];
    vec2i[0] = luaL_checkinteger(L, 3);
    vec2i[1] = luaL_checkinteger(L, 4);

    float v_speed = luaL_optnumber(L, 5, 1.0f);
    int v_min = luaL_optinteger(L, 6, 0);
    int v_max = luaL_optinteger(L, 7, 0);
    const char* format = luaL_optstring(L, 8, "%d");

    bool result = ImGui::DragInt(label, (int*)&vec2i, v_speed, v_min, v_max, format);

    lua_pushinteger(L, vec2i[0]);
    lua_pushinteger(L, vec2i[1]);
    lua_pushboolean(L, result);
    return 2;
}

int ImGui_impl_DragInt3(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    static int vec3i[3];
    vec3i[0] = luaL_checkinteger(L, 3);
    vec3i[1] = luaL_checkinteger(L, 4);
    vec3i[2] = luaL_checkinteger(L, 5);

    float v_speed = luaL_optnumber(L, 5, 1.0f);
    int v_min = luaL_optinteger(L, 6, 0);
    int v_max = luaL_optinteger(L, 7, 0);
    const char* format = luaL_optstring(L, 8, "%d");

    bool result = ImGui::DragInt(label, (int*)&vec3i, v_speed, v_min, v_max, format);

    lua_pushinteger(L, vec3i[0]);
    lua_pushinteger(L, vec3i[1]);
    lua_pushinteger(L, vec3i[2]);
    lua_pushboolean(L, result);
    return 2;
}

int ImGui_impl_DragInt4(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    static int vec4i[4];
    vec4i[0] = luaL_checkinteger(L, 3);
    vec4i[1] = luaL_checkinteger(L, 4);
    vec4i[2] = luaL_checkinteger(L, 5);
    vec4i[2] = luaL_checkinteger(L, 6);

    float v_speed = luaL_optnumber(L, 7, 1.0f);
    int v_min = luaL_optinteger(L, 8, 0);
    int v_max = luaL_optinteger(L, 9, 0);
    const char* format = luaL_optstring(L, 10, "%d");

    bool result = ImGui::DragInt(label, (int*)&vec4i, v_speed, v_min, v_max, format);

    lua_pushinteger(L, vec4i[0]);
    lua_pushinteger(L, vec4i[1]);
    lua_pushinteger(L, vec4i[2]);
    lua_pushinteger(L, vec4i[3]);
    lua_pushboolean(L, result);
    return 2;
}

int ImGui_impl_DragIntRange2(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    int v_current_min = luaL_checkinteger(L, 3);
    int v_current_max = luaL_checkinteger(L, 4);
    float v_speed = luaL_optnumber(L, 5, 1.0f);
    int v_min = luaL_optinteger(L, 6, 0.0f);
    int v_max = luaL_optinteger(L, 7, 0.0f);
    const char* format = luaL_optstring(L, 8, "%d");
    const char* format_max = luaL_optstring(L, 9, NULL);

    bool result = ImGui::DragIntRange2(label, &v_current_min, &v_current_max, v_speed, v_min, v_max, format, format_max);

    lua_pushinteger(L, v_current_min);
    lua_pushinteger(L, v_current_max);
    lua_pushboolean(L, result);
    return 3;
}

int ImGui_impl_DragScalar(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);
    ImGuiDataType data_type = luaL_checkinteger(L, 3);
    double value = luaL_checknumber(L, 4);
    float v_speed = luaL_checknumber(L, 5);
    double v_min = luaL_optnumber(L, 6, NULL);
    double v_max = luaL_optnumber(L, 7, NULL);
    const char* format = luaL_optstring(L, 8, NULL);
    float power = luaL_optnumber(L, 9, 1.0f);

    bool result = ImGui::DragScalar(label, data_type, (void *)&value, v_speed, (void *)&v_min, (void *)&v_max, format, power);

    lua_pushnumber(L, value);
    lua_pushboolean(L, result);
    return 2;
}

// Widgets: Sliders
//bool          SliderFloat(const char* label, float* v, float v_min, float v_max, const char* format = "%.3f", float power = 1.0f);
//bool          SliderFloat2(const char* label, float v[2], float v_min, float v_max, const char* format = "%.3f", float power = 1.0f);
//bool          SliderFloat3(const char* label, float v[3], float v_min, float v_max, const char* format = "%.3f", float power = 1.0f);
//bool          SliderFloat4(const char* label, float v[4], float v_min, float v_max, const char* format = "%.3f", float power = 1.0f);
//bool          SliderAngle(const char* label, float* v_rad, float v_degrees_min = -360.0f, float v_degrees_max = +360.0f, const char* format = "%.0f deg");
//bool          SliderInt(const char* label, int* v, int v_min, int v_max, const char* format = "%d");
//bool          SliderInt2(const char* label, int v[2], int v_min, int v_max, const char* format = "%d");
//bool          SliderInt3(const char* label, int v[3], int v_min, int v_max, const char* format = "%d");
//bool          SliderInt4(const char* label, int v[4], int v_min, int v_max, const char* format = "%d");
//bool          SliderScalar(const char* label, ImGuiDataType data_type, void* p_data, const void* p_min, const void* p_max, const char* format = NULL, float power = 1.0f);
//bool          SliderScalarN(const char* label, ImGuiDataType data_type, void* p_data, int components, const void* p_min, const void* p_max, const char* format = NULL, float power = 1.0f);
//bool          VSliderFloat(const char* label, const ImVec2& size, float* v, float v_min, float v_max, const char* format = "%.3f", float power = 1.0f);
//bool          VSliderInt(const char* label, const ImVec2& size, int* v, int v_min, int v_max, const char* format = "%d");
//bool          VSliderScalar(const char* label, const ImVec2& size, ImGuiDataType data_type, void* p_data, const void* p_min, const void* p_max, const char* format = NULL, float power = 1.0f);

// Widgets: Input with Keyboard
//bool          InputText(const char* label, char* buf, size_t buf_size, ImGuiInputTextFlags flags = 0, ImGuiInputTextCallback callback = NULL, void* user_data = NULL);
//bool          InputTextMultiline(const char* label, char* buf, size_t buf_size, const ImVec2& size = ImVec2(0, 0), ImGuiInputTextFlags flags = 0, ImGuiInputTextCallback callback = NULL, void* user_data = NULL);
//bool          InputTextWithHint(const char* label, const char* hint, char* buf, size_t buf_size, ImGuiInputTextFlags flags = 0, ImGuiInputTextCallback callback = NULL, void* user_data = NULL);
//bool          InputFloat(const char* label, float* v, float step = 0.0f, float step_fast = 0.0f, const char* format = "%.3f", ImGuiInputTextFlags flags = 0);
//bool          InputFloat2(const char* label, float v[2], const char* format = "%.3f", ImGuiInputTextFlags flags = 0);
//bool          InputFloat3(const char* label, float v[3], const char* format = "%.3f", ImGuiInputTextFlags flags = 0);
//bool          InputFloat4(const char* label, float v[4], const char* format = "%.3f", ImGuiInputTextFlags flags = 0);
//bool          InputInt(const char* label, int* v, int step = 1, int step_fast = 100, ImGuiInputTextFlags flags = 0);
//bool          InputInt2(const char* label, int v[2], ImGuiInputTextFlags flags = 0);
//bool          InputInt3(const char* label, int v[3], ImGuiInputTextFlags flags = 0);
//bool          InputInt4(const char* label, int v[4], ImGuiInputTextFlags flags = 0);
//bool          InputDouble(const char* label, double* v, double step = 0.0, double step_fast = 0.0, const char* format = "%.6f", ImGuiInputTextFlags flags = 0);
//bool          InputScalar(const char* label, ImGuiDataType data_type, void* p_data, const void* p_step = NULL, const void* p_step_fast = NULL, const char* format = NULL, ImGuiInputTextFlags flags = 0);
//bool          InputScalarN(const char* label, ImGuiDataType data_type, void* p_data, int components, const void* p_step = NULL, const void* p_step_fast = NULL, const char* format = NULL, ImGuiInputTextFlags flags = 0);

// Widgets: Color Editor/Picker (tip: the ColorEdit* functions have a little colored preview square that can be left-clicked to open a picker, and right-clicked to open an option menu.)
//bool          ColorEdit3(const char* label, float col[3], ImGuiColorEditFlags flags = 0);
//bool          ColorEdit4(const char* label, float col[4], ImGuiColorEditFlags flags = 0);
//bool          ColorPicker3(const char* label, float col[3], ImGuiColorEditFlags flags = 0);
//bool          ColorPicker4(const char* label, float col[4], ImGuiColorEditFlags flags = 0, const float* ref_col = NULL);
//bool          ColorButton(const char* desc_id, const ImVec4& col, ImGuiColorEditFlags flags = 0, ImVec2 size = ImVec2(0, 0));
//void          SetColorEditOptions(ImGuiColorEditFlags flags);

// Widgets: Trees
//bool          TreeNode(const char* label);
//bool          TreeNode(const char* str_id, const char* fmt, ...) IM_FMTARGS(2);   // helper variation to easily decorelate the id from the displayed string. Read the FAQ about why and how to use ID. to align arbitrary text at the same level as a TreeNode() you can use Bullet().
//bool          TreeNode(const void* ptr_id, const char* fmt, ...) IM_FMTARGS(2);   // "
//bool          TreeNodeV(const char* str_id, const char* fmt, va_list args) IM_FMTLIST(2);
//bool          TreeNodeV(const void* ptr_id, const char* fmt, va_list args) IM_FMTLIST(2);
//bool          TreeNodeEx(const char* label, ImGuiTreeNodeFlags flags = 0);
//bool          TreeNodeEx(const char* str_id, ImGuiTreeNodeFlags flags, const char* fmt, ...) IM_FMTARGS(3);
//bool          TreeNodeEx(const void* ptr_id, ImGuiTreeNodeFlags flags, const char* fmt, ...) IM_FMTARGS(3);
//bool          TreeNodeExV(const char* str_id, ImGuiTreeNodeFlags flags, const char* fmt, va_list args) IM_FMTLIST(3);
//bool          TreeNodeExV(const void* ptr_id, ImGuiTreeNodeFlags flags, const char* fmt, va_list args) IM_FMTLIST(3);
//void          TreePush(const char* str_id);                                       // ~ Indent()+PushId(). Already called by TreeNode() when returning true, but you can call TreePush/TreePop yourself if desired.
//void          TreePush(const void* ptr_id = NULL);                                // "
//void          TreePop();                                                          // ~ Unindent()+PopId()
//float         GetTreeNodeToLabelSpacing();                                        // horizontal distance preceding label when using TreeNode*() or Bullet() == (g.FontSize + style.FramePadding.x*2) for a regular unframed TreeNode
//bool          CollapsingHeader(const char* label, ImGuiTreeNodeFlags flags = 0);  // if returning 'true' the header is open. doesn't indent nor push on ID stack. user doesn't have to call TreePop().
//bool          CollapsingHeader(const char* label, bool* p_open, ImGuiTreeNodeFlags flags = 0); // when 'p_open' isn't NULL, display an additional small close button on upper right of the header
//void          SetNextItemOpen(bool is_open, ImGuiCond cond = 0);                  // set next TreeNode/CollapsingHeader open state.

// Widgets: Selectables
//bool          Selectable(const char* label, bool selected = false, ImGuiSelectableFlags flags = 0, const ImVec2& size = ImVec2(0, 0));
//bool          Selectable(const char* label, bool* p_selected, ImGuiSelectableFlags flags = 0, const ImVec2& size = ImVec2(0, 0));

// Widgets: List Boxes
//bool          ListBox(const char* label, int* current_item, const char* const items[], int items_count, int height_in_items = -1);
//bool          ListBox(const char* label, int* current_item, bool (*items_getter)(void* data, int idx, const char** out_text), void* data, int items_count, int height_in_items = -1);
//bool          ListBoxHeader(const char* label, const ImVec2& size = ImVec2(0, 0));
//bool          ListBoxHeader(const char* label, int items_count, int height_in_items = -1);
//void          ListBoxFooter();

// Widgets: Data Plotting
//void          PlotLines(const char* label, const float* values, int values_count, int values_offset = 0, const char* overlay_text = NULL, float scale_min = FLT_MAX, float scale_max = FLT_MAX, ImVec2 graph_size = ImVec2(0, 0), int stride = sizeof(float));
//void          PlotLines(const char* label, float(*values_getter)(void* data, int idx), void* data, int values_count, int values_offset = 0, const char* overlay_text = NULL, float scale_min = FLT_MAX, float scale_max = FLT_MAX, ImVec2 graph_size = ImVec2(0, 0));
//void          PlotHistogram(const char* label, const float* values, int values_count, int values_offset = 0, const char* overlay_text = NULL, float scale_min = FLT_MAX, float scale_max = FLT_MAX, ImVec2 graph_size = ImVec2(0, 0), int stride = sizeof(float));
//void          PlotHistogram(const char* label, float(*values_getter)(void* data, int idx), void* data, int values_count, int values_offset = 0, const char* overlay_text = NULL, float scale_min = FLT_MAX, float scale_max = FLT_MAX, ImVec2 graph_size = ImVec2(0, 0));

// Widgets: Value() Helpers.
int ImGui_impl_Value(lua_State *L)
{
    const char* prefix = luaL_checkstring(L, 2);
    const int valueType = lua_type(L, 3);
    switch(valueType)
    {
        case LUA_TBOOLEAN:
            {
                ImGui::Value(prefix, lua_toboolean(L, 3) > 0);
                break;
            }
        case LUA_TNUMBER:
            {
                float n = lua_tonumber(L, 3);
                int intN = (int)n;
                if (n == intN)
                {
                    ImGui::Value(prefix, intN);
                }
                else
                {
                    ImGui::Value(prefix, n, luaL_optstring(L, 4, NULL));
                }
                break;
            }
        default:
            {
                lua_pushstring(L, "Type mismatch. 'Number' or 'Boolean' expected.");
                lua_error(L);
                break;
            }
    }

    return 0;
}

// Widgets: Menus
//bool          BeginMenuBar();
//void          EndMenuBar();
//bool          BeginMainMenuBar();
//void          EndMainMenuBar();
//bool          BeginMenu(const char* label, bool enabled = true);
//void          EndMenu();
//bool          MenuItem(const char* label, const char* shortcut = NULL, bool selected = false, bool enabled = true);  // return true when activated. shortcuts are displayed for convenience but not processed by ImGui at the moment
//bool          MenuItem(const char* label, const char* shortcut, bool* p_selected, bool enabled = true);              // return true when activated + toggle (*p_selected) if p_selected != NULL

// Tooltips
//void          BeginTooltip();
//void          EndTooltip();
//void          SetTooltip(const char* fmt, ...)
//void          SetTooltipV(const char* fmt, va_list args)

// Popups, Modals
// Popups: begin/end functions
//bool          BeginPopup(const char* str_id, ImGuiWindowFlags flags = 0);
//bool          BeginPopupModal(const char* name, bool* p_open = NULL, ImGuiWindowFlags flags = 0);
//void          EndPopup();
// Popups: open/close functions
//void          OpenPopup(const char* str_id, ImGuiPopupFlags popup_flags = 0);
//bool          OpenPopupContextItem(const char* str_id = NULL, ImGuiPopupFlags popup_flags = 1);
//void          CloseCurrentPopup();
// Popups: open+begin combined functions helpers
//bool          BeginPopupContextItem(const char* str_id = NULL, ImGuiPopupFlags popup_flags = 1);
//bool          BeginPopupContextWindow(const char* str_id = NULL, ImGuiPopupFlags popup_flags = 1);
//bool          BeginPopupContextVoid(const char* str_id = NULL, ImGuiPopupFlags popup_flags = 1);
// Popups: test function
//bool          IsPopupOpen(const char* str_id, ImGuiPopupFlags flags = 0);

// Columns
//void          Columns(int count = 1, const char* id = NULL, bool border = true);
//void          NextColumn();
//int           GetColumnIndex();
//float         GetColumnWidth(int column_index = -1);
//void          SetColumnWidth(int column_index, float width);
//float         GetColumnOffset(int column_index = -1);
//void          SetColumnOffset(int column_index, float offset_x);
//int           GetColumnsCount();

// Tab Bars, Tabs
//bool          BeginTabBar(const char* str_id, ImGuiTabBarFlags flags = 0);
//void          EndTabBar();
//bool          BeginTabItem(const char* label, bool* p_open = NULL, ImGuiTabItemFlags flags = 0);
//void          EndTabItem();
//void          SetTabItemClosed(const char* tab_or_docked_window_label);

// Logging/Capture
//void          LogToTTY(int auto_open_depth = -1);
//void          LogToFile(int auto_open_depth = -1, const char* filename = NULL);
//void          LogToClipboard(int auto_open_depth = -1);
//void          LogFinish();
//void          LogButtons();
//void          LogText(const char* fmt, ...) IM_FMTARGS(1);

// Drag and Drop
//bool          BeginDragDropSource(ImGuiDragDropFlags flags = 0);
//bool          SetDragDropPayload(const char* type, const void* data, size_t sz, ImGuiCond cond = 0);
//void          EndDragDropSource();
//bool                  BeginDragDropTarget();
//const ImGuiPayload*   AcceptDragDropPayload(const char* type, ImGuiDragDropFlags flags = 0);
//void                  EndDragDropTarget();
//const ImGuiPayload*   GetDragDropPayload();

// Clipping
//void          PushClipRect(const ImVec2& clip_rect_min, const ImVec2& clip_rect_max, bool intersect_with_current_clip_rect);
//void          PopClipRect();

// Focus, Activation
//void          SetItemDefaultFocus();
//void          SetKeyboardFocusHere(int offset = 0);

// Item/Widgets Utilities
//bool          IsItemHovered(ImGuiHoveredFlags flags = 0);
//bool          IsItemActive();
//bool          IsItemFocused();
//bool          IsItemClicked(ImGuiMouseButton mouse_button = 0);
//bool          IsItemVisible();
//bool          IsItemEdited();
//bool          IsItemActivated();
//bool          IsItemDeactivated();
//bool          IsItemDeactivatedAfterEdit();
//bool          IsItemToggledOpen();
//bool          IsAnyItemHovered();
//bool          IsAnyItemActive();
//bool          IsAnyItemFocused();
//ImVec2        GetItemRectMin();
//ImVec2        GetItemRectMax();
//ImVec2        GetItemRectSize();
//void          SetItemAllowOverlap();

// Miscellaneous Utilities
//bool          IsRectVisible(const ImVec2& size);
//bool          IsRectVisible(const ImVec2& rect_min, const ImVec2& rect_max);
//double        GetTime();
//int           GetFrameCount();
//ImDrawList*   GetBackgroundDrawList();
//ImDrawList*   GetForegroundDrawList();
//ImDrawListSharedData* GetDrawListSharedData();
//const char*   GetStyleColorName(ImGuiCol idx);
//void          SetStateStorage(ImGuiStorage* storage);
//ImGuiStorage* GetStateStorage();
//void          CalcListClipping(int items_count, float items_height, int* out_items_display_start, int* out_items_display_end);
//bool          BeginChildFrame(ImGuiID id, const ImVec2& size, ImGuiWindowFlags flags = 0);
//void          EndChildFrame();

// Text Utilities
//ImVec2        CalcTextSize(const char* text, const char* text_end = NULL, bool hide_text_after_double_hash = false, float wrap_width = -1.0f);

// Inputs Utilities: Keyboard
//int           GetKeyIndex(ImGuiKey imgui_key);
//bool          IsKeyDown(int user_key_index);
//bool          IsKeyPressed(int user_key_index, bool repeat = true);
//bool          IsKeyReleased(int user_key_index);
//int           GetKeyPressedAmount(int key_index, float repeat_delay, float rate);
//void          CaptureKeyboardFromApp(bool want_capture_keyboard_value = true);

// Inputs Utilities: Mouse
//bool          IsMouseDown(ImGuiMouseButton button);
//bool          IsMouseClicked(ImGuiMouseButton button, bool repeat = false);
//bool          IsMouseReleased(ImGuiMouseButton button);
//bool          IsMouseDoubleClicked(ImGuiMouseButton button);
//bool          IsMouseHoveringRect(const ImVec2& r_min, const ImVec2& r_max, bool clip = true);
//bool          IsMousePosValid(const ImVec2* mouse_pos = NULL);
//bool          IsAnyMouseDown();
//ImVec2        GetMousePos();
//ImVec2        GetMousePosOnOpeningCurrentPopup();
//bool          IsMouseDragging(ImGuiMouseButton button, float lock_threshold = -1.0f);
//ImVec2        GetMouseDragDelta(ImGuiMouseButton button = 0, float lock_threshold = -1.0f);
//void          ResetMouseDragDelta(ImGuiMouseButton button = 0);
//ImGuiMouseCursor GetMouseCursor();
//void          SetMouseCursor(ImGuiMouseCursor cursor_type);
//void          CaptureMouseFromApp(bool want_capture_mouse_value = true);

/// STYLES

int ImGui_impl_StyleDark(lua_State *L)
{
    ImGui::StyleColorsDark();
    return 0;
}

int ImGui_impl_StyleLight(lua_State *L)
{
    ImGui::StyleColorsLight();
    return 0;
}

int ImGui_impl_StyleClassic(lua_State *L)
{
    ImGui::StyleColorsClassic();
    return 0;
}

/// DISPLAY SIZE

int ImGui_impl_SetDisplaySize(lua_State *L)
{
    float w = luaL_checknumber(L, 2);
    float h = luaL_checknumber(L, 3);

    ImGuiIO& io = ImGui::GetIO();

    io.DisplaySize.x = w;
    io.DisplaySize.y = h;

    return 0;
}

int ImGui_impl_SetDisplayScale(lua_State *L)
{
    float sx = luaL_checknumber(L, 2);
    float sy = luaL_checknumber(L, 3);

    ImGuiIO& io = ImGui::GetIO();

    io.DisplayFramebufferScale.x = sx;
    io.DisplayFramebufferScale.y = sy;

    return 0;
}

// Color Utilities
//ImVec4        ColorConvertU32ToFloat4(ImU32 in);
//ImU32         ColorConvertFloat4ToU32(const ImVec4& in);

int ImGui_impl_ColorConvertRGBtoHSV(lua_State *L)
{
    float r = luaL_checknumber(L, 2);
    float g = luaL_checknumber(L, 3);
    float b = luaL_checknumber(L, 4);

    float h = 0;
    float s = 0;
    float v = 0;

    ImGui::ColorConvertRGBtoHSV(r, g, b, h, s, v);

    lua_pushnumber(L, h);
    lua_pushnumber(L, s);
    lua_pushnumber(L, v);

    return 3;
}

int ImGui_impl_ColorConvertHSVtoRGB(lua_State *L)
{
    float h = luaL_checknumber(L, 2);
    float s = luaL_checknumber(L, 3);
    float v = luaL_checknumber(L, 4);

    float r = 0;
    float g = 0;
    float b = 0;

    ImGui::ColorConvertHSVtoRGB(h, s, v, r, g, b);

    lua_pushnumber(L, r);
    lua_pushnumber(L, g);
    lua_pushnumber(L, b);

    return 3;
}

////////////////////////////////////////////////////////////////////////////////
///
/// DEMOS
///
////////////////////////////////////////////////////////////////////////////////

int ImGui_impl_ShowUserGuide(lua_State *L)
{
    ImGui::ShowUserGuide();

    return 0;
}

int ImGui_impl_ShowDemoWindow(lua_State *L)
{
    ImGui::ShowDemoWindow();

    return 0;
}

int ImGui_impl_ShowAboutWindow(lua_State *L)
{
    ImGui::ShowAboutWindow();

    return 0;
}

int ImGui_impl_ShowStyleEditor(lua_State *L)
{
    ImGui::ShowStyleEditor();

    return 0;
}

int ImGui_impl_ShowFontSelector(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);

    ImGui::ShowFontSelector(label);

    return 0;
}

int ImGui_impl_ShowMetricsWindow(lua_State *L)
{
    ImGui::ShowMetricsWindow();

    return 0;
}

int ImGui_impl_ShowStyleSelector(lua_State *L)
{
    const char* label = luaL_checkstring(L, 2);

    bool open = ImGui::ShowStyleSelector(label);

    lua_pushboolean(L, open);

    return 1;
}

int ImGui_my_dumpTable(lua_State *L, int idx = 1)
{
    lua_pushnil(L);
    int i = 1;
    while(lua_next(L, idx - 2) != 0)
    {
        int t = lua_type(L, -1);
        switch(t)
        {
            case LUA_TSTRING:
                ImGui::Text("%d:`%s'", i, lua_tostring(L, -2));
                break;
            case LUA_TBOOLEAN:
                ImGui::Text("%d: %s", i,lua_toboolean(L, -2) ? "true" : "false");
                break;
            case LUA_TNUMBER:
                ImGui::Text("%d: %g",  i, lua_tonumber(L, -2));
                break;
            case LUA_TTABLE:
                ImGui::Text("%d: table {...}",  idx + i);
            default:
                ImGui::Text("%d: %s", i, lua_typename(L, -2));
            break;

            i++;
        }
        lua_pop(L, 1);
    }
    return 0;
}

int ImGui_my_ShowLuaStack(lua_State *L)
{
    if (!ImGui::Begin("LUA Stack Dump"))
    {
        ImGui::End();
        return 0;
    }

    int i=lua_gettop(L);
    while(i)
    {
        int t = lua_type(L, i);
        switch (t)
        {
            case LUA_TSTRING:
                ImGui::Text("%d:`%s'", i, lua_tostring(L, i));
                break;
            case LUA_TBOOLEAN:
                ImGui::Text("%d: %s",i,lua_toboolean(L, i) ? "true" : "false");
                break;
            case LUA_TNUMBER:
                ImGui::Text("%d: %g",  i, lua_tonumber(L, i));
                break;
            case LUA_TTABLE:
                if (ImGui::CollapsingHeader(lua_typename(L, t)))
                {
                    //ImGui_my_dumpTable(L, i);
                    ImGui::Text("%i", i);
                }
            default:
                ImGui::Text("* %d: %s", i, lua_typename(L, t));
            break;
        }
        i--;
    }

    ImGui::End();

    return 0;
}

int ImGui_my_test(lua_State *L)
{

    return 0;
}

int loader(lua_State *L)
{
    Binder binder(L);
    const luaL_Reg functionList[] =
    {
        {"test", ImGui_my_test},
        // Mouse inputs
        {"onMouseHover", ImGui_impl_MouseHover},
        {"onMouseMove", ImGui_impl_MouseMove},
        {"onMouseDown", ImGui_impl_MouseDown},
        {"onMouseUp", ImGui_impl_MouseUp},
        {"onMouseWheel", ImGui_impl_MouseWheel},

        // Touch inputs TODO

        // Keyboard
        {"onKeyUp", ImGui_impl_KeyUp},
        {"onKeyDown", ImGui_impl_KeyDown},
        {"onKeyChar", ImGui_impl_KeyChar},

        // Input flags
        {"wantCaptureMouse", ImGui_impl_WantCaptureMouse},
        {"wantCaptureKeyboard", ImGui_impl_WantCaptureKeyboard},
        {"wantTextInput", ImGui_impl_WantTextInput},

        // Colors
        {"colorConvertRGBtoHSV", ImGui_impl_ColorConvertRGBtoHSV},
        {"colorConvertHSVtoRGB", ImGui_impl_ColorConvertHSVtoRGB},

        // Styles
        {"setDarkStyle", ImGui_impl_StyleDark},
        {"setLightStyle", ImGui_impl_StyleLight},
        {"setClassicStyle", ImGui_impl_StyleClassic},

        // Childs
        {"beginChild", ImGui_impl_BeginChild},
        {"endChild", ImGui_impl_EndChild},

        {"isWindowAppearing", ImGui_impl_IsWindowAppearing},
        {"isWindowCollapsed", ImGui_impl_IsWindowCollapsed},
        {"isWindowFocused", ImGui_impl_IsWindowFocused},
        {"isWindowHovered", ImGui_impl_IsWindowHovered},
        {"getWindowPos", ImGui_impl_GetWindowPos},
        {"getWindowSize", ImGui_impl_GetWindowSize},
        {"getWindowWidth", ImGui_impl_GetWindowWidth},
        {"getWindowHeight", ImGui_impl_GetWindowHeight},

        {"setNextWindowPos", ImGui_impl_SetNextWindowPos},
        {"setNextWindowSize", ImGui_impl_SetNextWindowSize},
        {"setNextWindowContentSize", ImGui_impl_SetNextWindowContentSize},
        {"setNextWindowCollapsed", ImGui_impl_SetNextWindowCollapsed},
        {"setNextWindowFocus", ImGui_impl_SetNextWindowFocus},
        {"setNextWindowBgAlpha", ImGui_impl_SetNextWindowBgAlpha},
        {"setWindowPos", ImGui_impl_SetWindowPos},
        {"setWindowSize", ImGui_impl_SetWindowSize},
        {"setWindowCollapsed", ImGui_impl_SetWindowCollapsed},
        {"setWindowFocus", ImGui_impl_SetWindowFocus},

        {"getContentRegionMax", ImGui_impl_GetContentRegionMax},
        {"getContentRegionAvail", ImGui_impl_GetContentRegionAvail},
        {"getWindowContentRegionMin", ImGui_impl_GetWindowContentRegionMin},
        {"getWindowContentRegionMax", ImGui_impl_GetWindowContentRegionMax},
        {"getWindowContentRegionWidth", ImGui_impl_GetWindowContentRegionWidth},

        {"getScrollX", ImGui_impl_GetScrollX},
        {"getScrollY", ImGui_impl_GetScrollY},
        {"getScrollMaxX", ImGui_impl_GetScrollMaxX},
        {"getScrollMaxY", ImGui_impl_GetScrollMaxY},
        {"setScrollX", ImGui_impl_SetScrollX},
        {"setScrollY", ImGui_impl_SetScrollY},
        {"setScrollHereX", ImGui_impl_SetScrollHereX},
        {"setScrollHereY", ImGui_impl_SetScrollHereY},
        {"setScrollFromPosX", ImGui_impl_SetScrollFromPosX},
        {"setScrollFromPosY", ImGui_impl_SetScrollFromPosY},

        {"pushStyleColor", ImGui_impl_PushStyleColor},
        {"popStyleColor", ImGui_impl_PopStyleColor},
        {"pushStyleVar", ImGui_impl_PushStyleVar},
        {"popStyleVar", ImGui_impl_PopStyleVar},
        {"getStyleColorVec4", ImGui_impl_GetStyleColorVec4},
        {"getColorU32", ImGui_impl_GetColorU32},

        {"pushItemWidth", ImGui_impl_PushItemWidth},
        {"popItemWidth", ImGui_impl_PopItemWidth},
        {"setNextItemWidth", ImGui_impl_SetNextItemWidth},
        {"calcItemWidth", ImGui_impl_CalcItemWidth},
        {"pushTextWrapPos", ImGui_impl_PushTextWrapPos},
        {"popTextWrapPos", ImGui_impl_PopTextWrapPos},
        {"pushAllowKeyboardFocus", ImGui_impl_PushAllowKeyboardFocus},
        {"popAllowKeyboardFocus", ImGui_impl_PopAllowKeyboardFocus},
        {"pushButtonRepeat", ImGui_impl_PushButtonRepeat},
        {"popButtonRepeat", ImGui_impl_PopButtonRepeat},

        {"separator", ImGui_impl_Separator},
        {"sameLine", ImGui_impl_SameLine},
        {"newLine", ImGui_impl_NewLine},
        {"spacing", ImGui_impl_Spacing},
        {"dummy", ImGui_impl_Dummy},
        {"indent", ImGui_impl_Indent},
        {"unindent", ImGui_impl_Unindent},
        {"beginGroup", ImGui_impl_BeginGroup},
        {"endGroup", ImGui_impl_EndGroup},

        {"getCursorPos", ImGui_impl_GetCursorPos},
        {"getCursorPosX", ImGui_impl_GetCursorPosX},
        {"getCursorPosY", ImGui_impl_GetCursorPosY},
        {"getCursorStartPos", ImGui_impl_GetCursorStartPos},
        {"getCursorScreenPos", ImGui_impl_GetCursorScreenPos},
        {"alignTextToFramePadding", ImGui_impl_AlignTextToFramePadding},
        {"getTextLineHeight", ImGui_impl_GetTextLineHeight},
        {"getTextLineHeightWithSpacing", ImGui_impl_GetTextLineHeightWithSpacing},
        {"getFrameHeight", ImGui_impl_GetFrameHeight},
        {"getFrameHeightWithSpacing", ImGui_impl_GetFrameHeightWithSpacing},

        {"pushSID", ImGui_impl_PushID_string},
        {"pushNID", ImGui_impl_PushID_number},
        {"popID", ImGui_impl_PopID},
        {"getSID", ImGui_impl_GetID_string},

        {"textUnformatted", ImGui_impl_TextUnformatted},
        {"text", ImGui_impl_Text},
        {"textColored", ImGui_impl_TextColored},
        {"textDisabled", ImGui_impl_TextDisabled},
        {"textWrapped", ImGui_impl_TextWrapped},
        {"labelText", ImGui_impl_LabelText},
        {"bulletText", ImGui_impl_BulletText},

        {"button", ImGui_impl_Button},
        {"smallButton", ImGui_impl_SmallButton},
        {"invisibleButton", ImGui_impl_InvisibleButton},
        {"arrowButton", ImGui_impl_ArrowButton},
        {"image", ImGui_impl_Image},
        {"imageButton", ImGui_impl_ImageButton},
        {"checkbox", ImGui_impl_Checkbox},
        {"checkboxFlags", ImGui_impl_CheckboxFlags},
        {"radioButton", ImGui_impl_RadioButton},
        {"progressBar", ImGui_impl_ProgressBar},
        {"bullet", ImGui_impl_Bullet},
        {"beginCombo", ImGui_impl_BeginCombo},
        {"endCombo", ImGui_impl_EndCombo},
        {"combo", ImGui_impl_Combo},

        {"dragFloat", ImGui_impl_DragFloat},
        {"dragFloat2", ImGui_impl_DragFloat2},
        {"dragFloat3", ImGui_impl_DragFloat3},
        {"dragFloat4", ImGui_impl_DragFloat4},

        {"dragInt", ImGui_impl_DragInt},
        {"dragInt2", ImGui_impl_DragInt2},
        {"dragInt3", ImGui_impl_DragInt3},
        {"dragInt4", ImGui_impl_DragInt4},
        {"dragIntRange2", ImGui_impl_DragIntRange2},
        {"dragScalar", ImGui_impl_DragScalar},

        // Windows
        {"beginWindow", ImGui_impl_Begin},
        {"endWindow", ImGui_impl_End},

        // Render
        {"newFrame", ImGui_impl_NewFrame},
        {"render", ImGui_impl_Render},
        {"endFrame", ImGui_impl_EndFrame},

        // Display size
        {"setDisplaySize", ImGui_impl_SetDisplaySize},
        {"setDisplayScale", ImGui_impl_SetDisplayScale},

        // Demos
        {"showUserGuide", ImGui_impl_ShowUserGuide},
        {"showDemoWindow", ImGui_impl_ShowDemoWindow},
        {"showAboutWindow", ImGui_impl_ShowAboutWindow},
        {"showStyleEditor", ImGui_impl_ShowStyleEditor},
        {"showFontSelector", ImGui_impl_ShowFontSelector},
        {"showMetricsWindow", ImGui_impl_ShowMetricsWindow},
        {"showStyleSelector", ImGui_impl_ShowStyleSelector},
        {"showLuaStack", ImGui_my_ShowLuaStack},
        {NULL, NULL}
    };

    binder.createClass("ImGui", "Sprite", initImGui, destroyImGui, functionList);

    luaL_newweaktable(L);
    luaL_rawsetptr(L, LUA_REGISTRYINDEX, &keyWeak);


    // TODO:
    // Enums
    lua_getglobal(L, "ImGui");
    bindWindowFlagsEnum(L);
    bindInputTextFlagsEnum(L);
    bindTreeNodeFlagsEnum(L);
    bindPopupFlagsEnum(L);
    bindSelectableFlagsEnum(L);
    bindComboFlagsEnum(L);
    bindTabBarFlagsEnum(L);
    bindTabItemFlagsEnum(L);
    bindFocusedFlagsEnum(L);
    bindHoveredFlagsEnum(L);

    lua_pushstring(L, ImGui::GetVersion());
    lua_setfield(L, -2, "_VERSION");
    lua_pop(L, 1);

    return 1;
}

static void g_initializePlugin(lua_State *L)
{
    ::L = L;
    lua_getglobal(L, "package");
    lua_getfield(L, -1, "preload");

    lua_pushcfunction(L, loader);
    lua_setfield(L, -2, "ImGui");

    lua_pop(L, 2);

}

static void g_deinitializePlugin(lua_State *_UNUSED(L)) {  }

REGISTER_PLUGIN_NAMED("ImGui", "1.0.0", imgui)
